import pytest
from unittest.mock import patch
from utils import initialize_fit
import torch
from irtorch.estimation_algorithms import AEIRT
from irtorch.models import MonotoneNN


# The @pytest.fixture decorator is used to create fixture methods.
# This method is called once per test method that uses it, and its return value is passed to the test method as an argument.
# pytest.fixture with params creates two fixtures, one for the CPU and one for the GPU.
# The ids parameter is used to give the tests meaningful names
class TestAEIRT:
    @pytest.fixture()
    def algorithm(self, device, latent_variables, item_categories, data_loaders):
        if device == "cuda" and not torch.cuda.is_available():
            pytest.skip("GPU is not available.")

        model = MonotoneNN(
            latent_variables = latent_variables,
            item_categories = item_categories,
            hidden_dim = [3]
        )
        algorithm = AEIRT(
            model=model,
            hidden_layers_encoder=[20],  # 1 hidden layer with 20 neurons
            nonlinear_encoder=torch.nn.ELU()
        )

        algorithm.imputation_method = "zero"
        # Set DataLoaders from the fixture
        algorithm.data_loader, algorithm.validation_data_loader = data_loaders

        initialize_fit(algorithm)
        return algorithm

    def test_forward(self, algorithm: AEIRT, test_data):
        output = algorithm.forward(test_data)
        assert output.shape == (120, algorithm.model.max_item_responses * algorithm.model.items)

    def test_latent_scores(self, algorithm: AEIRT, test_data):
        output = algorithm.z_scores(test_data.to(next(algorithm.parameters()).device))
        assert output.shape == (120, algorithm.model.latent_variables)

    def test__train_step(self, algorithm: AEIRT):
        previous_loss = float("inf")
        for epoch in range(2):
            loss = algorithm._train_step(epoch)
            assert loss <= previous_loss  # Loss should decrease
            previous_loss = loss

    def test__validation_step(self, algorithm: AEIRT):
        log_likelihood = algorithm._validation_step()
        assert isinstance(log_likelihood, float)
        assert log_likelihood > 0

    def test__impute_missing_zero(self, algorithm: AEIRT):
        a, b = 5, 5
        data = torch.full((a, b), 5)
        no_missing_mask = torch.full((a, b), 0)
        missing_mask = torch.tensor(
            [
                [0, 0, 1, 0, 0],
                [0, 1, 0, 0, 0],
                [0, 0, 0, 0, 1],
                [0, 0, 0, 0, 0],
                [1, 1, 0, 0, 0],
            ]
        )

        algorithm.imputation_method = "zero"
        imputed_data = algorithm._impute_missing(data, missing_mask)
        assert torch.equal(
            imputed_data,
            torch.tensor(
                [
                    [5, 5, 0, 5, 5],
                    [5, 0, 5, 5, 5],
                    [5, 5, 5, 5, 0],
                    [5, 5, 5, 5, 5],
                    [0, 0, 5, 5, 5],
                ]
            ),
        )
        imputed_data = algorithm._impute_missing(data, no_missing_mask)
        assert torch.equal(imputed_data, data)

    # The following is a test for the fit function of the AEIRTNeuralNet class
    def test_fit(self, algorithm: AEIRT, test_data):
        # Mock the inner functions that would be called during training
        with patch.object(
            algorithm, "_train_step", return_value=torch.tensor(0.5)
        ) as mocked_train_step, patch.object(
            algorithm, "_validation_step", return_value=torch.tensor(0.5)
        ) as mocked_validation_step:
            # Call fit function
            algorithm.fit(
                train_data=test_data[0:100],
                validation_data=test_data[100:120],
                max_epochs=5
            )

            # Check if inner functions are called
            assert mocked_train_step.call_count == 5
            assert mocked_validation_step.call_count == 5



import pytest
import torch
from irtorch.load_dataset import swedish_national_mathematics_2, swedish_national_mathematics_1, swedish_sat_verbal_2022, swedish_sat_quantitative_2022, swedish_sat_2022, swedish_sat_2022_binary, big_five

def test_swedish_national_mathematics_2():
    data = swedish_national_mathematics_2()
    assert data.dtype == torch.float32
    assert data.shape == (1401, 28)

def test_swedish_national_mathematics_1():
    data = swedish_national_mathematics_1()
    assert data.dtype == torch.float32
    assert data.shape == (1008, 28)

# TODO add back after adding SAT to package
# def test_swedish_sat_verbal_2022():
#     data, correct_category = swedish_sat_verbal_2022()
#     assert data.dtype == torch.float32
#     assert data.shape == (38068, 80)
#     assert len(correct_category) == 80
#     assert all([isinstance(x, int) for x in correct_category])

# def test_swedish_sat_quantitative_2022():
#     data, correct_category = swedish_sat_quantitative_2022()
#     assert data.dtype == torch.float32
#     assert data.shape == (38068, 80)
#     assert len(correct_category) == 80
#     assert all([isinstance(x, int) for x in correct_category])

# def test_swedish_sat_2022():
#     data, correct_category = swedish_sat_2022()
#     assert data.dtype == torch.float32
#     assert data.shape == (38068, 160)
#     assert len(correct_category) == 160
#     assert all([isinstance(x, int) for x in correct_category])

# def test_swedish_sat_2022_binary():
#     data = swedish_sat_2022_binary()
#     assert data.dtype == torch.float32
#     assert data.shape == (38068, 160)
#     assert data[~data.isnan()].max() == 1.0
#     assert data[~data.isnan()].min() == 0.0

def test_big_five():
    data = big_five()
    assert data.dtype == torch.float32
    assert data.shape == (19719, 50)
    assert data[~data.isnan()].max() == 4.0
    assert data[~data.isnan()].min() == 0.0



import pytest
import torch
import pandas as pd
from irtorch.models import Parametric

def test_1pl_forward():
    model = Parametric(
        latent_variables = 2,
        item_categories=[2, 2, 2],
        model = "1PL",
        item_z_relationships=torch.tensor([[True, True], [True, True], [False, True]])
    )
    original_weights = model.weight_param.clone()
    original_bias = model.bias_param.clone()

    optimizer = torch.optim.Adam(
        [{"params": model.parameters()}], lr=0.02, amsgrad=True
    )
    z = torch.tensor([[0.1, 0.2], [0.3, 0.4], [0.5, 0.5], [0.5, 0.6]])
    data = torch.tensor([[0, 1, 0], [0, 0, 1], [1, 1, 1], [1, 1, 1]]).float()
    for _ in range(2): # update two times with the same data
        optimizer.zero_grad()
        output = model.forward(z)
        assert output.shape == (4, 6), "Incorrect output shape"

        loss = -model.log_likelihood(data=data, output=output)
        # Compute gradients
        loss.backward()
        optimizer.step()

    assert torch.all(original_weights != model.weight_param), "Free weights should have changed"
    assert torch.all(original_bias != model.bias_param), "Free bias should have changed"


def test_2pl_forward():
    model = Parametric(
        latent_variables = 2,
        model = "2PL",
        item_categories=[2, 2, 2],
        item_z_relationships=torch.tensor([[True, True], [True, True], [False, True]])
    )
    original_weights = model.weight_param.clone()
    original_bias = model.bias_param.clone()

    optimizer = torch.optim.Adam(
        [{"params": model.parameters()}], lr=0.02, amsgrad=True
    )
    z = torch.tensor([[0.1, 0.2], [0.3, 0.4], [0.5, 0.5], [0.5, 0.6]])
    data = torch.tensor([[0, 1, 0], [0, 1, 1], [1, 1, 1], [1, 1, 1]]).float()
    for _ in range(2): # update two times with the same data
        optimizer.zero_grad()
        output = model.forward(z)
        assert output.shape == (4, 6), "Incorrect output shape"

        loss = -model.log_likelihood(data=data, output=output)
        # Compute gradients
        loss.backward()
        optimizer.step()

    assert torch.all(original_weights != model.weight_param), "Free weights should have changed"
    assert torch.all(original_bias != model.bias_param), "Free bias should have changed"

def test_gpc_forward():
    model = Parametric(
        latent_variables = 2,
        model = "GPC",
        item_categories=[2, 3, 3],
        item_z_relationships=torch.tensor([[True, True], [True, True], [False, True]])
    )
    original_weights = model.weight_param.clone()
    original_bias = model.bias_param.clone()

    optimizer = torch.optim.Adam(
        [{"params": model.parameters()}], lr=0.02, amsgrad=True
    )
    z = torch.tensor([[0.1, 0.2], [0.3, 0.4], [0.5, 0.5], [0.8, 0.6]])
    data = torch.tensor([[0, 1, 0], [0, 1, 1], [1, 2, 1], [1, 2, 2]]).float()
    for _ in range(2): # update two times with the same data
        optimizer.zero_grad()
        output = model.forward(z)
        assert output.shape == (4, 9), "Incorrect output shape"

        loss = -model.log_likelihood(data=data, output=output)
        # Compute gradients
        loss.backward()
        optimizer.step()

    assert torch.all(original_weights != model.weight_param), "Free weights should have changed"
    assert torch.all(original_bias != model.bias_param), "Free bias should have changed"

def test_nominal_forward():
    # With reference category
    model = Parametric(
        latent_variables = 2,
        model = "nominal",
        item_categories=[2, 3, 3],
        item_z_relationships=torch.tensor([[True, True], [True, True], [False, True]]),
        reference_category=True
    )
    original_weights = model.weight_param.clone()
    original_bias = model.bias_param.clone()

    optimizer = torch.optim.Adam(
        [{"params": model.parameters()}], lr=0.02, amsgrad=True
    )
    z = torch.tensor([[0.1, 0.2], [0.3, 0.4], [0.5, 0.5], [0.8, 0.6]])
    data = torch.tensor([[0, 1, 0], [0, 1, 1], [1, 2, 1], [1, 2, 2]]).float()
    for _ in range(2): # update two times with the same data
        optimizer.zero_grad()
        output = model.forward(z)
        assert output.shape == (4, 9), "Incorrect output shape"

        loss = -model.log_likelihood(data=data, output=output)
        # Compute gradients
        loss.backward()
        optimizer.step()

    assert model.weight_param.shape == torch.Size([8]), "Incorrect weight shape"
    assert model.bias_param.shape == torch.Size([5]), "Incorrect weight shape"
    assert torch.all(original_weights != model.weight_param), "Free weights should have changed"
    assert torch.all(original_bias != model.bias_param), "Free bias should have changed"

    # Without reference category
    model = Parametric(
        latent_variables = 2,
        model = "nominal",
        item_categories=[2, 3, 3],
        item_z_relationships=torch.tensor([[True, True], [True, True], [False, True]]),
        reference_category=False
    )
    original_weights = model.weight_param.clone()
    original_bias = model.bias_param.clone()

    optimizer = torch.optim.Adam(
        [{"params": model.parameters()}], lr=0.02, amsgrad=True
    )
    z = torch.tensor([[0.1, 0.2], [0.3, 0.4], [0.5, 0.5], [0.8, 0.6]])
    data = torch.tensor([[0, 1, 0], [0, 1, 1], [1, 2, 1], [1, 2, 2]]).float()
    for _ in range(2): # update two times with the same data
        optimizer.zero_grad()
        output = model.forward(z)
        assert output.shape == (4, 9), "Incorrect output shape"

        loss = -model.log_likelihood(data=data, output=output)
        # Compute gradients
        loss.backward()
        optimizer.step()

    assert model.weight_param.shape == torch.Size([13]), "Incorrect weight shape"
    assert model.bias_param.shape == torch.Size([8]), "Incorrect weight shape"
    assert torch.all(original_weights != model.weight_param), "Free weights should have changed"
    assert torch.all(original_bias != model.bias_param), "Free bias should have changed"
    
def test_probabilities_from_output():
    model = Parametric(
        latent_variables = 2,
        model = "GPC",
        item_categories=[2, 3, 4],
        item_z_relationships=torch.tensor([[True, True], [True, True], [False, True]])
    )

    optimizer = torch.optim.Adam(
        [{"params": model.parameters()}], lr=0.02, amsgrad=True
    )
    z = torch.tensor([[0.1, 0.2], [0.3, 0.4], [0.5, 0.5], [0.8, 0.6]])
    data = torch.tensor([[0, 1, 0], [0, 1, 1], [1, 2, 1], [1, 2, 3]]).float()
    for _ in range(2): # update two times with the same data
        optimizer.zero_grad()
        output = model.forward(z)
        assert output.shape == (4, 12), "Incorrect output shape"

        loss = -model.log_likelihood(data=data, output=output)
        # Compute gradients
        loss.backward()
        optimizer.step()

    probabilities = model.probabilities_from_output(model(z))
    assert probabilities.shape == (4, 3, 4), "Incorrect probabilities shape"
    assert torch.all(probabilities[:, 0, 2:4] == 0.0), "probabilities for missing categories should be 0"
    assert torch.all(probabilities[:, 1, 3:4] == 0.0), "probabilities for missing categories should be 0"
    assert torch.allclose(probabilities.sum(dim=2), torch.ones(probabilities.shape[0], probabilities.shape[1])), "probabilities for missing categories should be 0"

def test_item_parameters(latent_variables):
    for model_type in ["nominal", "1PL", "2PL", "GPC"]:
        if model_type in ["GPC", "nominal"]:
            item_categories = [2, 3, 4]
        else:
            item_categories = [2, 2, 2]
        model = Parametric(
            latent_variables = latent_variables,
            model = model_type,
            item_categories=item_categories,
            item_z_relationships=torch.tensor([[True], [True], [True]]).repeat(1, latent_variables)
        )
        parameters = model.item_parameters(IRT_format=False)

        assert isinstance(parameters, pd.DataFrame), f"Output should be a DataFrame for {model_type} model"
        if model_type == "nominal":
            assert parameters.shape == (model.items, (model.latent_variables + 1) * model.max_item_responses), f"Incorrect shape of parameters DataFrame for {model_type} model"
        else:
            assert parameters.shape == (model.items, model.latent_variables + model.max_item_responses), f"Incorrect shape of parameters DataFrame for {model_type} model"

        parameters_irt = model.item_parameters(IRT_format=True)

        assert isinstance(parameters_irt, pd.DataFrame), f"Output should be a DataFrame for {model_type} model"
        if model_type == "nominal":
            assert parameters_irt.shape == (model.items, (model.latent_variables + 1) * model.max_item_responses), f"Incorrect shape of parameters DataFrame for {model_type} model"
        elif model.latent_variables == 1:
            assert parameters_irt.shape == (model.items, model.latent_variables + model.max_item_responses - 1), f"Incorrect shape of parameters DataFrame for {model_type} model"
        else:
            assert parameters_irt.shape == (model.items, model.latent_variables + model.max_item_responses), f"Incorrect shape of parameters DataFrame for {model_type} model"

def test_item_z_relationship_directions():
    model = Parametric(
        latent_variables = 2,
        model = "GPC",
        item_categories=[2, 3, 4],
        item_z_relationships=torch.tensor([[True, True], [True, True], [False, True]])
    )
    model.weight_param.data = torch.tensor([0.5, 0.6, -0.8, -1, 2])

    directions = model.item_z_relationship_directions()
    assert directions.shape == (3, 2), "Incorrect directions shape"
    assert torch.all(directions[0, :] == 1), "Incorrect directions"
    assert torch.all(directions[1, :] == -1), "Incorrect directions"
    assert directions[2, 0] == 0, "Incorrect directions"
    assert directions[2, 1] == 1, "Incorrect directions"

    model = Parametric(
        latent_variables = 2,
        model = "nominal",
        item_categories=[2, 3, 3],
        item_z_relationships=torch.tensor([[True, True], [True, True], [False, True]]),
        mc_correct=[2, 2, 3],
        reference_category=True
    )

    model.weight_param.data = torch.tensor([1.0, -1.0, -1.0, -1.0, -1.0, 3.0, 1.0, 1.0])

    directions = model.item_z_relationship_directions()
    assert directions.shape == (3, 2), "Incorrect directions shape"
    assert directions[0, 0] == 1, "Incorrect directions"
    assert directions[0, 1] == -1, "Incorrect directions"
    assert directions[1, 0] == -1, "Incorrect directions"
    assert directions[1, 1] == -1, "Incorrect directions"
    assert directions[2, 0] == 0, "Incorrect directions"
    assert directions[2, 1] == 1, "Incorrect directions"

    # Without reference category
    model = Parametric(
        latent_variables = 2,
        model = "nominal",
        item_categories=[2, 3, 3],
        item_z_relationships=torch.tensor([[True, True], [True, True], [False, True]]),
        mc_correct=[2, 2, 3],
        reference_category=False
    )

    model.weight_param.data = torch.tensor([0.0, 0.0, 1.0, -1.0, 0.0, 0.0, -1.0, -1.0, -1.0, 3.0, 0.0, 1.0, 1.0])

    directions = model.item_z_relationship_directions()
    assert directions.shape == (3, 2), "Incorrect directions shape"
    assert directions[0, 0] == 1, "Incorrect directions"
    assert directions[0, 1] == -1, "Incorrect directions"
    assert directions[1, 0] == -1, "Incorrect directions"
    assert directions[1, 1] == -1, "Incorrect directions"
    assert directions[2, 0] == 0, "Incorrect directions"
    assert directions[2, 1] == 1, "Incorrect directions"



import logging
from abc import ABC, abstractmethod
import torch
from irtorch.models import BaseIRTModel

logger = logging.getLogger('irtorch')

class BaseIRTAlgorithm(ABC):
    """
    Abstract base class for IRT algorithms. All IRT algorithms should inherit from this class.

    Parameters
    ----------
    model : BaseIRTModel, optional
        The model to train. Needs to inherit irtorch.models.BaseIRTModel.
    one_hot_encoded : bool, optional
        Whether the algorithm uses one-hot encoded data. (default is False)
    """
    def __init__(
        self,
        model: BaseIRTModel,
        one_hot_encoded: bool = False,
    ):
        super().__init__()
        self.model = model
        self.imputation_method = "zero"
        self.train_data = None

        if model.mc_correct is not None:
            self.one_hot_encoded = True
        else:
            self.one_hot_encoded = one_hot_encoded

    @abstractmethod
    def fit(self, train_data: torch.Tensor, **kwargs):
        """
        Fit the model to the data.

        Parameters
        ----------
        train_data : torch.Tensor
            The training data.
        """
        # Store all training data within the model instance
        if self.train_data is not None:
            self.train_data = torch.cat((self.train_data, train_data), dim=0).contiguous()
        else:
            self.train_data = train_data.contiguous()

    @abstractmethod
    def _impute_missing_with_prior(self, batch, missing_mask):
        """
        Impute missing values with the prior.

        Parameters
        ----------
        batch : torch.Tensor
            The batch of data.
        missing_mask : torch.Tensor
            The mask of missing values.

        Returns
        -------
        torch.Tensor
            The imputed data.
        """
        raise NotImplementedError(
            "Prior imputation not implemented for this algorithm."
        )

    def fix_missing_values(self, data: torch.Tensor):
        """
        Deal with missing values so that the data can be used for fitting.

        Parameters
        ----------
        data : torch.Tensor
            The data to fix.

        Returns
        -------
        torch.Tensor
            The data with missing values imputed.
        """
        if data.isnan().any():
            data[data.isnan()] = -1

        if self.model.model_missing:
            data = data + 1 # handled in z_scores for nn
        else:
            if self.imputation_method == "zero":
                data[data == -1] = 0
            else:
                # self._impute_missing(data, data.isnan())
                # see also helper_function impute_missing
                raise NotImplementedError("Imputation methods not implemented yet")

        return data



import logging
import torch
import copy
import pandas as pd
from itertools import product
import torch.multiprocessing as mp
from irtorch.irt import IRT

logger = logging.getLogger('irtorch')

def cross_validation(
    irt_model: IRT,
    data: torch.Tensor,
    folds: int,
    params_grid: dict,
    device: str = "cuda" if torch.cuda.is_available() else "cpu",
    z_estimation_method: str = "ML",
    **kwargs
) -> pd.DataFrame:
    """
    Perform cross-validation on the given model and data. Uses log-likelihood for model evaluation. Note that for running on the CPU on windows, `if __name__ == "__main__":` needs to be added to the main script before calling this function, see examples.

    Parameters
    ----------
    irt_model : IRT
        The irt model to train. Note that this should be an untrained model.
    data : torch.Tensor
        The data to use for cross-validation. The data is randomly shuffled before splitting into folds.
    folds : int
        The number of folds to use for cross-validation.
    params_grid : dict
        The hyperparameters to use for cross-validation. All need to be arguments for the model fit method.
    z_estimation_method : str, optional
        Method used to obtain the z scores. Also used for bit scores as they require the z scores. Can be 'NN', 'ML', 'EAP' or 'MAP' for neural network, maximum likelihood, expected a posteriori or maximum a posteriori respectively. (default is 'ML')
    device : str, optional
        The device to use for training. Can be 'cpu' for CPU or 'cuda' for GPU (if available). The default is 'cuda' if a GPU is available, otherwise 'cpu'.
    **kwargs
        Additional keyword arguments to pass to the model fit method.

    Returns
    -------
    list
        A list of dictionaries containing the hyperparameters and the corresponding cross-validation scores.

    Examples
    --------
    This example demonstrates how to use cross_validation() function with the Swedish National Mathematics dataset.

    First, we import necessary modules, load the data and split it into a training and testing set:

    >>> from irtorch import IRT
    >>> from irtorch.load_dataset import swedish_national_mathematics_2
    >>> from irtorch.utils import split_data
    >>> from irtorch.cross_validation import cross_validation
    >>> data_math = swedish_national_mathematics_2()
    >>> train_data, test_data = split_data(data_math, 0.8)

    Next, we initialize the IRT model:

    >>> irt_model = IRT(data=data_math)

    We then set up a grid of parameters for cross-validation:

    >>> params_grid = {
    ...     'learning_rate': [0.05, 0.1],
    ...     'batch_size': [64, 128],
    ... }

    Finally, we perform cross-validation to find a good set of parameters:

    >>> if __name__ == "__main__":
    ...     result = cross_validation(irt_model, data=train_data, folds=5, params_grid=params_grid, z_estimation_method="NN", device="cpu")
    """
    if device == "cuda" and not torch.cuda.is_available():
        raise ValueError("CUDA is not available on this machine, use device = 'cpu'.")

    # randomly scrable the data
    data = data[torch.randperm(data.shape[0])]
    data_folds = torch.chunk(data, folds, dim=0)

    param_combinations = list(product(*params_grid.values()))
    param_comb_names = list(params_grid.keys())
    param_dicts = [
        {name: value for name, value in zip(param_comb_names, combination)}
        for combination in param_combinations
    ]
    # Add additional kwargs to each parameter dictionary
    for param_dict in param_dicts:
        param_dict.update(kwargs)

    logger.info(f"Performing cross-validation with {len(param_dicts)} parameter combinations")

    # Prepare arguments for multiprocessing
    jobs = []
    for params in param_dicts:
        for fold in range(folds):
            train_data = torch.cat([data_folds[i] for i in range(folds) if i != fold])
            validation_data = data_folds[fold]
            jobs.append((copy.deepcopy(irt_model), train_data, validation_data, params, z_estimation_method, device))

    if device == "cpu":
        cores_to_use = mp.cpu_count()
        print(f"Using {cores_to_use} cores")
        original_threads = torch.get_num_threads()
        try:
            with mp.Pool(cores_to_use) as pool:
                results = pool.starmap(_cv_fold, jobs)
        finally:
            torch.set_num_threads(original_threads)

    elif device == "cuda":
        results = []
        for job in jobs:
            results.append(_cv_fold(*job))

    # average over folds
    results = pd.DataFrame(results)
    results.drop(list(kwargs.keys()), axis=1, inplace=True)
    results = results.groupby(param_comb_names).mean().reset_index()
    return results

def _cv_fold(irt_model : IRT, train_data, validation_data, params, z_estimation_method, device):
    if device == "cpu":
        torch.set_num_threads(1) # One thread per core, to avoid overloading the CPU

    irt_model.fit(train_data, device=device, **params)
    log_likelihood = irt_model.log_likelihood(validation_data, z_estimation_method = z_estimation_method, reduction="sum").item()

    return {**params, "log_likelihood": log_likelihood}



import pytest
import torch
from irtorch.layers import NegationLayer, SoftplusLinear

def test_SoftplusLinear_forward():
    zero_outputs=torch.tensor([1, 1, 1, 0, 0, 0, 0, 1, 1])
    sp_linear = SoftplusLinear(
        in_features=2,
        out_features=9,
        zero_outputs=zero_outputs,
    )

    original_weights = sp_linear.raw_weight_param.clone()
    original_bias = sp_linear.bias_param.clone()

    # Each row increases in all inputs, so the output should increase over rows
    optimizer = torch.optim.Adam(
        [{"params": sp_linear.parameters()}], lr=0.02, amsgrad=True
    )
    for _ in range(2):
        optimizer.zero_grad()
        result = sp_linear(torch.tensor([[-1.0, 0.1], [0.1, 1.0], [1.0, 2.0]]))
        assert result.shape == (3, 9), "Incorrect output shape"
        assert torch.all((result[1:] - result[:-1])[:, sp_linear.free_bias.bool()] > 0), "Not all updated columns are strictly increasing over rows."
        assert torch.all(result[:, zero_outputs] == 0.0)
        
        result.sum().backward()
        assert torch.all(sp_linear.bias_param.grad != 0) , "Not all bias gradients are computed."
        assert torch.all(sp_linear.raw_weight_param.grad != 0) , "Not all weight gradients are computed."
        optimizer.step()

    assert torch.all(original_weights != sp_linear.raw_weight_param), "Free weights should have changed"
    assert torch.all(original_bias != sp_linear.bias_param), "Free weights should have changed"

def test_SoftplusLinear_forward_with_separate_groups():
    zero_inputs=torch.tensor([0, 0, 0, 0, 1, 0])
    zero_outputs=torch.tensor([1, 1, 1, 0, 0, 0, 0, 1, 1])
    sp_linear = SoftplusLinear(
        in_features=6,
        out_features=9,
        separate_inputs=torch.tensor([1, 3, 2]),
        separate_outputs=torch.tensor([2, 2, 5]),
        zero_inputs=zero_inputs,
        zero_outputs=zero_outputs,
    )

    original_weights = sp_linear.raw_weight_param.clone()
    original_bias = sp_linear.bias_param.clone()

    # Each row increases in all inputs, so the output should increase over rows
    optimizer = torch.optim.Adam(
        [{"params": sp_linear.parameters()}], lr=0.02, amsgrad=True
    )
    for _ in range(2):
        optimizer.zero_grad()
        result = sp_linear(torch.tensor([[-1.0, -0.4, 0.3, 0.6, 1.0, 1.1],
                                         [0.1, 0.2, 0.4, 0.7, 1.2, 1.3],
                                         [0.4, 0.3, 4.2, 6.1, 8.0, 8.5]]))
        assert result.shape == (3, 9), "Incorrect output shape"
        assert torch.all((result[1:] - result[:-1])[:, sp_linear.free_bias.bool()] > 0), "Not all updated columns are strictly increasing over rows."
        assert torch.all(result[:, zero_outputs] == 0.0)
        
        result.sum().backward()
        assert torch.all(sp_linear.bias_param.grad != 0) , "Not all bias gradients are computed."
        assert torch.all(sp_linear.raw_weight_param.grad != 0) , "Not all weight gradients are computed."
        optimizer.step()

    assert torch.all(original_weights != sp_linear.raw_weight_param), "Free weights should have changed"
    assert torch.all(original_bias != sp_linear.bias_param), "Free weights should have changed"


def test_NegationLayer_forward():
    neg_layer = NegationLayer(
        item_z_relationships=torch.tensor([1, 0, 1]),
        inputs_per_items=3,
        zero_outputs=torch.tensor([0, 0, 1, 0, 0, 0, 0, 0, 1]).bool(),
    )
    with torch.no_grad():
        neg_layer.weight_param[0].fill_(2)
        neg_layer.weight_param[1].fill_(-1)

    # neg_layer.weights
    input = torch.tensor([[2.0] * 9, [1.0] * 9, [-1.0] * 9])
    input[:, 8] = 0.0
    input[:, 2:6] = 0.0
    result = neg_layer(input)

    result.sum().backward()
    assert torch.all(neg_layer.weight_param.grad != 0) , "Not all weight gradients are computed."
    assert torch.equal(result, torch.tensor([
        [ 4.,  4., 0., 0.,  0.,  0., -2., -2., 0.],
        [ 2.,  2., 0., 0.,  0.,  0.,  -1.,  -1., 0.],
        [-2., -2., 0., 0.,  0.,  0.,  1.,  1., 0.]
    ]))

def test_NegationLayer_all_item_weights():
    neg_layer = NegationLayer(
        item_z_relationships=torch.tensor([1, 0, 1]),
        inputs_per_items=3,
        zero_outputs=torch.tensor([0, 0, 1, 0, 0, 0, 0, 0, 1]).bool(),
    )
    with torch.no_grad():
        neg_layer.weight_param[0].fill_(2)
        neg_layer.weight_param[1].fill_(-1)
        
    assert neg_layer.all_item_weights().equal(torch.tensor([ 2.,  0., -1.]))


def test_separate_weights():
    separate_inputs = torch.tensor([3, 2])
    separate_outputs = torch.tensor([3, 2])
    model = SoftplusLinear(5, 5)
    result = model.separate_weights(model.in_features, model.out_features, separate_inputs, separate_outputs)
    assert result.shape == (model.out_features, model.in_features)
    assert torch.all(result[0:3, 0:3] == True)
    assert torch.all(result[3:5, 3:5] == True)
    assert torch.all(result[0:3, 3:5] == False)
    assert torch.all(result[3:5, 0:3] == False)


import torch
import pytest
from irtorch.gaussian_mixture_torch import GaussianMixtureTorch


def test_gaussian_mixture_torch():
    # Create an instance of GaussianMixtureTorch
    gmm = GaussianMixtureTorch()

    # Generate some example data
    data = torch.cat(
        [
            torch.normal(mean=-3, std=1, size=(100, 1)),
            torch.normal(mean=3, std=1, size=(100, 1)),
        ]
    )

    # Test the fit method
    gmm.fit(data)
    assert gmm.n_components is not None
    assert gmm.gmm is not None

    # Test the pdf method
    new_data = torch.tensor([[2.5]])  # a single observation
    density = gmm.pdf(new_data)
    assert torch.is_tensor(density)
    assert density is not None
    assert density > 0

    # Test the error handling in the pdf method
    gmm_empty = GaussianMixtureTorch()
    with pytest.raises(AttributeError):
        gmm_empty.pdf(new_data)



from abc import ABC, abstractmethod
import logging
import torch
from torch import nn
import torch.nn.functional as F
from irtorch._internal_utils import linear_regression

logger = logging.getLogger('irtorch')

class BaseIRTModel(ABC, nn.Module):
    """
    Abstract base class for Item Response Theory models. All IRT models should inherit from this class.
    
    Parameters
    ----------
    latent_variables : int
        The number of latent variables.
    item_categories : list[int]
        A list of the number of categories for each item.
    mc_correct : list[int], optional
        A list of the correct response category for each multiple choice item. (default is None)
    model_missing : bool, optional
        Whether to model missing data. (default is False)
    """
    def __init__(
        self,
        latent_variables: int,
        item_categories: list[int],
        mc_correct: list[int] = None,
        model_missing: bool = False
    ):
        super().__init__()
        if mc_correct is not None:
            if not all(item >= mc for item, mc in zip(item_categories, mc_correct)):
                raise ValueError("mc_correct cannot be greater than the number of categories for each item.")
        self.latent_variables = latent_variables
        self.item_categories = item_categories
        self.modeled_item_responses = [x + 1 for x in item_categories] if model_missing else item_categories
        self.max_item_responses = max(self.modeled_item_responses)
        self.items = len(item_categories)
        self.mc_correct = mc_correct
        self.model_missing = model_missing


    @abstractmethod
    def forward(self, z) -> torch.Tensor:
        """
        Forward pass of the model. PyTorch requires this method to be implemented in subclasses of nn.Module.

        Parameters
        ----------
        z : torch.Tensor
            Input tensor.

        Returns
        -------
        output : torch.Tensor
            Output tensor.
        """

    @abstractmethod
    def probabilities_from_output(self, output: torch.Tensor) -> torch.Tensor:
        """
        Compute probabilities from the output tensor from the forward method.

        Parameters
        ----------
        output : torch.Tensor
            Output tensor from the forward pass.

        Returns
        -------
        torch.Tensor
            3D tensor with dimensions (respondents, items, item categories)
        """

    def item_probabilities(self, z: torch.Tensor) -> torch.Tensor:
        """
        Returns the probabilities for each possible response for all items.

        Parameters
        ----------
        z : torch.Tensor
            A 2D tensor. Rows are respondents and latent variables are columns.

        Returns
        -------
        torch.Tensor
            3D tensor with dimensions (respondents, items, item categories)
        """
        output = self(z)
        return self.probabilities_from_output(output)


    def log_likelihood(
        self,
        data: torch.Tensor,
        output: torch.Tensor,
        loss_reduction: str = "sum",
    ) -> torch.Tensor:
        """
        Compute the log likelihood of the data given the model. This is equivalent to the negative cross entropy loss.

        Parameters
        ----------
        data: torch.Tensor
            A 2D tensor with test data. Columns are items and rows are respondents
        output: torch.Tensor
            A 2D tensor with output. Columns are item response categories and rows are respondents
        loss_reduction: str, optional 
            The reduction argument for torch.nn.CrossEntropyLoss. (default is 'sum')
        
        Returns
        -------
        torch.Tensor
            The log likelihood.
        """
        data = data.long()
        data = data.view(-1)
        reshaped_output = output.reshape(-1, self.max_item_responses)
        return -F.cross_entropy(reshaped_output, data, reduction=loss_reduction)

    def expected_item_sum_score(self, z: torch.Tensor, return_item_scores: bool = True) -> torch.Tensor:
        """
        Computes the model expected item scores/sum scores for each respondent.

        Parameters
        ----------
        z : torch.Tensor
            A 2D tensor with z scores. Each row represents one respondent, and each column represents a latent variable.
        return_item_scores : bool, optional
            Whether to return the expected item scores. If False, returns the expected sum scores (default is True)

        Returns
        -------
        torch.Tensor
            A tensor with the expected scores for each respondent.
        """
        item_probabilities = self.item_probabilities(z)
        if self.mc_correct:
            item_scores = torch.zeros(item_probabilities.shape[1], item_probabilities.shape[2])
            item_scores.scatter_(1, (torch.tensor(self.mc_correct) - 1 + self.model_missing).unsqueeze(1), 1)
        else:
            item_scores = (torch.arange(item_probabilities.shape[2])).repeat(item_probabilities.shape[1], 1)
            if self.model_missing:
                item_scores[:, 1:] = item_scores[:, 1:] - 1
        expected_item_scores = (item_probabilities * item_scores).sum(dim=2)

        if return_item_scores:
            return expected_item_scores
        else:
            return expected_item_scores.sum(dim=1)
        
    def item_z_relationship_directions(self, z: torch.Tensor) -> torch.Tensor:
        """
        Get the relationships between each item and latent variable for a fitted model.

        Parameters
        ----------
        z : torch.Tensor
            A 2D tensor with latent z scores from the population of interest. Each row represents one respondent, and each column represents a latent variable.

        Returns
        -------
        torch.Tensor
            A 2D tensor with the relationships between the items and latent variables. Items are rows and latent variables are columns.
        """
        item_sum_scores = self.expected_item_sum_score(z)
        item_z_mask = torch.zeros((len(self.items), self.latent_variables)).bool()
        for item, _ in enumerate(self.modeled_item_responses):
            weights = linear_regression(z, item_sum_scores[:,item].reshape(-1, 1))[1:].reshape(-1)
            item_z_mask[item, :] = weights >= 0
        
        return item_z_mask.int()

    @torch.inference_mode()
    def sample_test_data(self, z: torch.Tensor) -> torch.Tensor:
        """
        Sample test data for the provided z scores.

        Parameters
        ----------
        z : torch.Tensor
            The latent scores.

        Returns
        -------
        torch.Tensor
            The sampled test data.
        """
        probs = self.item_probabilities(z)
        dist = torch.distributions.Categorical(probs)
        return dist.sample().float()
    
    @torch.inference_mode(False)
    def probability_gradients(self, z: torch.Tensor) -> torch.Tensor:
        """
        Calculate the gradients of the item response probabilities with respect to the z scores.

        Parameters
        ----------
        z : torch.Tensor
            A 2D tensor containing latent variable z scores. Each column represents one latent variable.

        Returns
        -------
        torch.Tensor
            A torch tensor with the gradients for each z score. Dimensions are (z rows, items, item categories, latent variables).
        """
        z = z.clone().requires_grad_(True)

        def compute_jacobian(z_single):
            jacobian = torch.func.jacrev(self.item_probabilities)(z_single.view(1, -1))
            return jacobian.squeeze((0, 3))

        logger.info("Computing Jacobian for all items and item categories...")
        # vectorized version of jacobian
        gradients = torch.vmap(compute_jacobian)(z)
        return gradients



import logging
from .irt import IRT

__version__ = '0.0.1'

logging.basicConfig(level=logging.INFO,
                    format='%(levelname)s: %(message)s',
                    datefmt='%Y-%m-%d %H:%M:%S')

logger = logging.getLogger('irtorch')



import pytest
import torch
from torch.nn.functional import softmax
from irtorch.models import BaseIRTModel
from unittest.mock import MagicMock, patch

class ConcreteIRTModel(BaseIRTModel):
    """
    Concrete implementation of BaseIRTModel for testing purposes.
    """
    def forward(self, z):
        pass

    def probabilities_from_output(self, output: torch.Tensor):
        pass

@pytest.fixture
def base_irt_model(latent_variables):
    return ConcreteIRTModel(latent_variables, [3, 4])


def test_expected_item_sum_score(base_irt_model: BaseIRTModel):
    # Create a mock for item_probabilities() method
    def item_probabilities_mock(*args, **kwargs):
        return torch.tensor([
            [[0.1, 0.1, 0.8, 0.0], [0.1, 0.1, 0.1, 0.7]],
            [[0.3, 0.3, 0.4, 0.0], [0.3, 0.5, 0.1, 0.1]]
        ])
    base_irt_model.item_probabilities = MagicMock(
        side_effect=item_probabilities_mock
    )

    expected_item_scores = base_irt_model.expected_item_sum_score(torch.randn((2, 2)), return_item_scores=True)
    assert torch.allclose(expected_item_scores, torch.tensor([[1.7, 2.4], [1.1, 1.0]]))
    assert torch.allclose(expected_item_scores.sum(dim=1), base_irt_model.expected_item_sum_score(torch.randn((2, 2)), return_item_scores=False))

    base_irt_model.model_missing = True
    expected_item_scores = base_irt_model.expected_item_sum_score(torch.randn((2, 2)), return_item_scores=True)
    assert torch.allclose(expected_item_scores, torch.tensor([[0.8, 1.5], [0.4, 0.3]]))
    assert torch.allclose(expected_item_scores.sum(dim=1), base_irt_model.expected_item_sum_score(torch.randn((2, 2)), return_item_scores=False))

    base_irt_model.model_missing = False
    base_irt_model.mc_correct = [1, 2]
    expected_item_scores = base_irt_model.expected_item_sum_score(torch.randn((2, 2)), return_item_scores=True)
    assert torch.allclose(expected_item_scores, torch.tensor([[0.1, 0.1], [0.3, 0.5]]))
    assert torch.allclose(expected_item_scores.sum(dim=1), base_irt_model.expected_item_sum_score(torch.randn((2, 2)), return_item_scores=False))
    
    base_irt_model.model_missing = True
    base_irt_model.mc_correct = [0, 1] # should give the same as above even if in practice invalid mc_correct
    expected_item_scores = base_irt_model.expected_item_sum_score(torch.randn((2, 2)), return_item_scores=True)
    assert torch.allclose(expected_item_scores, torch.tensor([[0.1, 0.1], [0.3, 0.5]]))
    assert torch.allclose(expected_item_scores.sum(dim=1), base_irt_model.expected_item_sum_score(torch.randn((2, 2)), return_item_scores=False))

    # what if we just have 1 respondent?
    def item_probabilities_mock2(*args, **kwargs):
        return torch.tensor([[[0.1, 0.1, 0.8, 0.0], [0.1, 0.1, 0.2, 0.7]]])
    
    base_irt_model.item_probabilities = MagicMock(
        side_effect=item_probabilities_mock2
    )
    base_irt_model.mc_correct = None
    base_irt_model.model_missing = False
    expected_item_scores = base_irt_model.expected_item_sum_score(torch.randn((1, 2)), return_item_scores=True)
    assert torch.allclose(expected_item_scores, torch.tensor([[1.7, 2.6]]))
    base_irt_model.mc_correct = [3, 1]
    base_irt_model.model_missing = False
    expected_item_scores = base_irt_model.expected_item_sum_score(torch.randn((1, 2)), return_item_scores=True)
    assert torch.allclose(expected_item_scores, torch.tensor([[0.8, 0.1]]))
    base_irt_model.mc_correct = None
    base_irt_model.model_missing = True
    expected_item_scores = base_irt_model.expected_item_sum_score(torch.randn((1, 2)), return_item_scores=True)
    assert torch.allclose(expected_item_scores, torch.tensor([[0.8, 1.6]]))

def test_probability_gradients(base_irt_model: BaseIRTModel):
    # Create a mock for item_probabilities() method
    def item_probabilities_mock(z):
        logits = torch.tensor([[[1, 2, 3, 0], [4, 3, 2, 1]]]).expand(1, 2, 4) * z.sum(dim=1).reshape(-1, 1, 1)
        logits[:, 0, 3] = -torch.inf
        return softmax(logits, dim=2)
    
    base_irt_model.item_probabilities = MagicMock(
        side_effect=item_probabilities_mock
    )

    input_z = torch.tensor([[-2.0, -3.0], [1.0, 2.0], [1.0, 1.0]])
    prob_gradients = base_irt_model.probability_gradients(input_z)

    assert prob_gradients.shape == (3, 2, 4, 2)

def test_sample_test_data(base_irt_model: BaseIRTModel):
    # Create a mock for item_probabilities() method
    def item_probabilities_mock(*args, **kwargs):
        return torch.tensor([
            [[0.1, 0.1, 0.8, 0.0], [0.1, 0.1, 0.1, 0.7]],
            [[0.3, 0.3, 0.4, 0.0], [0.3, 0.5, 0.1, 0.1]]
        ])
    base_irt_model.item_probabilities = MagicMock(side_effect=item_probabilities_mock)

    # Call sample_test_data
    z = torch.randn((2, 2))
    sampled_data = base_irt_model.sample_test_data(z)

    # Check the output shape and type
    assert sampled_data.shape == z.shape
    assert sampled_data.dtype == torch.float32

    # Check if the values are in the correct range (0 to number of categories - 1)
    assert torch.all((sampled_data >= 0) & (sampled_data < 4))



import logging
import pandas as pd
import torch
import torch.nn as nn
import torch.nn.functional as F
from irtorch.models.base_irt_model import BaseIRTModel

logger = logging.getLogger('irtorch')

class Parametric(BaseIRTModel):
    """
    Parametric IRT model. Currently supports 1PL, 2PL, 3PL, Generalized partial credit and nominal models.

    Parameters
    ----------
    latent_variables : int
        Number of latent variables.
    item_categories : list[int]
        Number of categories for each item. One integer for each item. Missing responses exluded.
    model : str
        Type of parametric model. Can be "1PL", "2PL", "3PL", "GPC" or "nominal".
    model_missing : bool, optional
        Whether to model missing item responses as separate item response categories. (Default: False)
    mc_correct : list[int], optional
        The correct response category for each item. If specified, the logits for the correct responses are cumulative logits. (Default: None)
    item_z_relationships : torch.Tensor, optional
        A boolean tensor of shape (items, latent_variables). If specified, the model will have connections between latent dimensions and items where the tensor is True. If left out, all latent variables and items are related (Default: None)
    reference_category : bool, optional
        Only for the nomimal model. Whether to use the first category as an unparameterized reference category. (Default: False and uses the original parameterization given by Bock(1972))
    """
    def __init__(
        self,
        latent_variables: int,
        item_categories: list[int],
        model: str,
        model_missing: bool = False,
        mc_correct: list[int] = None,
        item_z_relationships: torch.Tensor = None,
        reference_category: bool = False
    ):
        super().__init__(latent_variables=latent_variables, item_categories=item_categories, model_missing=model_missing, mc_correct=mc_correct)
        if item_z_relationships is not None:
            if item_z_relationships.shape != (len(item_categories), latent_variables):
                raise ValueError(
                    f"latent_item_connections must have shape ({len(item_categories)}, {latent_variables})."
                )
            assert(item_z_relationships.dtype == torch.bool), "latent_item_connections must be boolean type."
            assert(torch.all(item_z_relationships.sum(dim=1) > 0)), "all items must have a relationship with a least one latent variable."
        if model not in ["1PL", "2PL", "3PL", "GPC", "nominal"]:
            raise ValueError("model_type must be one of '1PL', '2PL', '3PL', 'GPC', 'nominal'.")
        if model in ["1PL", "2PL", "3PL"]:
            item_categories = [2] * len(item_categories)

        self.model = model
        self.output_size = self.items * self.max_item_responses

        if model == "1PL":
            free_weights = torch.ones(latent_variables)
        elif model in ["2PL", "3PL"]:
            free_weights = torch.zeros(self.items, self.max_item_responses - 1, latent_variables)
            for item, item_cat in enumerate(self.modeled_item_responses):
                if item_z_relationships is not None:
                    free_weights[item, 0:item_cat, :] = item_z_relationships[item, :]
                else:
                    free_weights[item, 0:item_cat, :] = 1.0
            free_weights = free_weights.reshape(-1, latent_variables)
        elif model == "nominal":
            free_weights = torch.zeros(self.items, self.max_item_responses, latent_variables)
            for item, item_cat in enumerate(self.modeled_item_responses):
                start_1 = 1 if reference_category else 0
                if item_z_relationships is not None:
                    free_weights[item, start_1:item_cat, :] = item_z_relationships[item, :]
                else:
                    free_weights[item, start_1:item_cat, :] = 1.0
            free_weights = free_weights.reshape(-1, latent_variables)
        elif model == "GPC":
            free_weights = torch.ones(self.items, latent_variables)
            self.register_buffer("gpc_weight_multiplier", torch.arange(0, self.max_item_responses).repeat(self.items))
            if item_z_relationships is not None:
                for item, item_cat in enumerate(self.modeled_item_responses):
                    free_weights[item, :] = item_z_relationships[item, :]

        if model == "1PL":
            self.weight_param = nn.Parameter(torch.zeros(latent_variables))
        elif model in ["2PL", "3PL", "nominal", "GPC"]:
            self.weight_param = nn.Parameter(torch.zeros(free_weights.sum().int()))
            if model == "3PL":
                self.guessing_param = nn.Parameter(torch.zeros(self.items))

        number_of_bias_parameters = sum(self.modeled_item_responses) if model == "nominal" and not reference_category else sum(self.modeled_item_responses) - self.items
        self.bias_param = nn.Parameter(torch.zeros(number_of_bias_parameters))
        first_category = torch.zeros(self.items, self.max_item_responses)
        if model != "nominal" or reference_category:
            first_category[:, 0] = 1.0
        first_category = first_category.reshape(-1)
        missing_category = torch.zeros(self.items, self.max_item_responses)
        for item, item_cat in enumerate(self.modeled_item_responses):
            missing_category[item, item_cat:self.max_item_responses] = 1.0
        missing_category = missing_category.reshape(-1)
        free_bias = (1 - first_category) * (1 - missing_category)
        self.register_buffer("free_weights", free_weights.bool())
        self.register_buffer("free_bias", free_bias.bool())
        self.register_buffer("missing_category", missing_category.bool())
        self.register_buffer("first_category", first_category.bool())
        self.reset_parameters()

    def reset_parameters(self) -> None:
        nn.init.normal_(self.weight_param, mean=1., std=0.01)
        nn.init.zeros_(self.bias_param)
        if self.model == "3PL":
            self.guessing_param.data.fill_(0.25)
    
    def forward(self, z: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the model.

        Parameters
        ----------
        z : torch.Tensor
            2D tensor with latent variables. Rows are respondents and latent variables are columns. 

        Returns
        -------
        output : torch.Tensor
            2D tensor. Rows are respondents and columns are item category logits.
        """
        bias = torch.zeros(self.output_size, device=z.device)
        bias[self.free_bias] = self.bias_param
        
        if self.model == "1PL":
            weighted_z = (self.weight_param * z).sum(dim=1, keepdim=True)
        elif self.model in ["2PL", "3PL", "GPC"]:
            weights = torch.zeros(self.items, self.latent_variables, device=z.device)
            weights[self.free_weights] = self.weight_param
            weighted_z = torch.matmul(z, weights.T).repeat_interleave(self.max_item_responses, dim=1)
            if self.model == "GPC":
                weighted_z *= self.gpc_weight_multiplier
        elif self.model == "nominal":
            weights = torch.zeros(self.output_size, self.latent_variables, device=z.device)
            weights[self.free_weights] = self.weight_param
            weighted_z = torch.matmul(z, weights.T)

        output = weighted_z + bias
        if self.model in ["GPC", "nominal"]:
            # stop gradients from flowing through the missing categories
            output[:, self.missing_category] = -torch.inf

        output[:, self.first_category] = 0

        return output

    def log_likelihood(
        self,
        data: torch.Tensor,
        output: torch.Tensor,
        loss_reduction: str = "sum",
    ) -> torch.Tensor:
        """
        Compute the log likelihood of the data given the model. This is equivalent to the negative cross entropy loss.

        Parameters
        ----------
        data: torch.Tensor
            A 2D tensor with test data. Columns are items and rows are respondents
        output: torch.Tensor
            A 2D tensor with output. Columns are item response categories and rows are respondents
        loss_reduction: str, optional 
            The reduction argument for torch.nn.CrossEntropyLoss. (default is 'sum')
        
        Returns
        -------
        torch.Tensor
            The log likelihood.
        """
        data = data.long()
        data = data.view(-1)
        reshaped_output = output.reshape(-1, self.max_item_responses)
        if self.model == "3PL":
            raise NotImplementedError("3PL not implemented yet")
            # probabilities = torch.softmax(reshaped_output, dim=1)
            # guessing = self.guessing_param.unsqueeze(1).repeat(int(data.shape[0]/self.items), reshaped_output.shape[1])
            # guessing[:, 0] *= -1
            # probabilities += guessing
            # return -F.nll_loss(torch.log(probabilities), data, reduction=loss_reduction)
        else:
            return -F.cross_entropy(reshaped_output, data, reduction=loss_reduction)

    def probabilities_from_output(self, output: torch.Tensor) -> list:
        """
        Compute item probabilities from the output tensor.

        Parameters
        ----------
        output : torch.Tensor
            Output tensor from the forward pass.

        Returns
        -------
        probabilities : torch.Tensor
            3D torch tensor with dimensions (respondents, items, item categories).
        """
        reshaped_output = output.reshape(-1, self.max_item_responses)
        return F.softmax(reshaped_output, dim=1).reshape(output.shape[0], self.items, self.max_item_responses)

    def item_parameters(self, IRT_format = False) -> pd.DataFrame:
        """
        Get the item parameters for a fitted model.

        Parameters
        ----------
        IRT_format : bool, optional
            Only for unidimensional models. Whether to return the item parameters in traditional IRT format. Otherwise returns weights and biases. (default is False)

        Returns
        -------
        pd.DataFrame
            A dataframe with the item parameters.
        """
        biases = torch.zeros(self.output_size)
        biases[self.free_bias] = self.bias_param
        biases = biases.reshape(-1, self.max_item_responses)
        if self.model == "1PL":
            weights = self.weight_param.repeat(self.items, 1)
        elif self.model == "nominal":
            weights = torch.zeros(self.output_size, self.latent_variables)
            weights[self.free_weights] = self.weight_param
            weights = weights.reshape(-1, self.max_item_responses * self.latent_variables)
        else:
            weights = torch.zeros(self.items, self.latent_variables)
            weights[self.free_weights] = self.weight_param

        weights_df = pd.DataFrame(weights.detach().numpy())
        if IRT_format and self.latent_variables == 1:
            if self.model == "nominal":
                weights_df.columns = [f"a{i+1}{j+1}" for i in range(self.latent_variables) for j in range(int(weights.shape[1]/self.latent_variables))]
                biases_df = pd.DataFrame(-(weights*biases).detach().numpy())
            else:
                weights_df.columns = [f"a{i+1}" for i in range(weights.shape[1])]
                biases_df = pd.DataFrame(-(weights*biases).detach()[:, 1:].numpy())
        else:
            if self.model == "nominal":
                weights_df.columns = [f"w{i+1}{j+1}" for i in range(self.latent_variables) for j in range(int(weights.shape[1]/self.latent_variables))]
            else:
                weights_df.columns = [f"w{i+1}" for i in range(weights.shape[1])]
            biases_df = pd.DataFrame(biases.detach().numpy())
            
        biases_df.columns = [f"b{i+1}" for i in range(biases_df.shape[1])]
        parameters = pd.concat([weights_df, biases_df], axis=1)

        return parameters

    @torch.inference_mode()
    def item_z_relationship_directions(self, *args) -> torch.Tensor:
        """
        Get the relationship directions between each item and latent variable for a fitted model.

        Returns
        -------
        torch.Tensor
            A 2D tensor with the relationships between the items and latent variables. Items are rows and latent variables are columns.
        """
        if self.model == "1PL":
            weights = self.weight_param.repeat(self.items, 1)
        if self.model == "nominal":
            model_weights = torch.zeros(self.output_size, self.latent_variables)
            model_weights[self.free_weights] = self.weight_param
            weights = torch.zeros(self.items, self.latent_variables)
            for item in range(self.items):
                weights[item] = model_weights[item * self.max_item_responses + self.mc_correct[item]-1]
        else:
            weights = torch.zeros(self.items, self.latent_variables)
            weights[self.free_weights] = self.weight_param
        return weights.sign().int()
    


import logging
import torch
from irtorch.models import BaseIRTModel
from irtorch.estimation_algorithms import BaseIRTAlgorithm, VAEIRT
from irtorch.irt_scorer import QuantileMVNormal, GaussianMixtureTorch
from irtorch.irt_scorer import IRTScorer
from irtorch._internal_utils import (
    impute_missing,
    conditional_score_distribution,
    sum_incorrect_probabilities,
)

logger = logging.getLogger('irtorch')

class IRTEvaluator:
    def __init__(self, model: BaseIRTModel, algorithm: BaseIRTAlgorithm, scorer: IRTScorer):
        """
        Initializes the IRTEvaluator class.

        Parameters
        ----------
        model : BaseIRTModel
            BaseIRTModel object.
        algorithm : BaseIRTAlgorithm
            BaseIRTAlgorithm object.
        scorer : IRTScorer
            IRTScorer object used to obtain latent variable scores.
        """
        self.model = model
        self.algorithm = algorithm
        self.scorer = scorer

    def _evaluate_data_z_input(
            self,
            data: torch.Tensor = None,
            z: torch.Tensor = None,
            z_estimation_method: str = "ML",
            ml_map_device: str = "cuda" if torch.cuda.is_available() else "cpu",
            lbfgs_learning_rate: float = 0.3,
        ):
        """"
        Helper function for evaluating the data and z inputs for various performance measure methods.

        Parameters
        ----------
        data : torch.Tensor
            The input data.
        z: torch.Tensor, optional
            The latent variable z scores for the provided data. If not provided, they will be computed using z_estimation_method.
        z_estimation_method : str, optional
            Method used to obtain the z scores. Can be 'NN', 'ML', 'EAP' or 'MAP' for neural network, maximum likelihood, expected a posteriori or maximum a posteriori respectively.
        ml_map_device: str, optional
            For ML and MAP. The device to use for computation. Can be 'cpu' or 'cuda'. (default is "cuda" if available else "cpu")
        lbfgs_learning_rate: float, optional
            For ML and MAP. The learning rate to use for the LBFGS optimizer. (default is 0.3)
        """
        if data is None:
            data = self.algorithm.train_data
        else:
            data = data.contiguous()

        if not self.model.model_missing:
            data = impute_missing(data, self.model.mc_correct, self.model.item_categories)

        if z is None:
            z = self.scorer.latent_scores(data=data, scale="z", z_estimation_method=z_estimation_method, ml_map_device=ml_map_device, lbfgs_learning_rate=lbfgs_learning_rate)

        data = self.algorithm.fix_missing_values(data)
        
        return data, z

    @torch.inference_mode()
    def residuals(
        self,
        data: torch.Tensor = None,
        z: torch.Tensor = None,
        z_estimation_method: str = "ML",
        average_over: str = "none",
    ) -> torch.Tensor:
        """
        Calculate the residuals of the model for the supplied data. 
        
        For multiple choice models, the residuals are computed as 1 - the probability of the selected response option.
        For other models, the residuals are computed as the difference between the observed and model expected item scores.

        Parameters
        ----------
        data : torch.Tensor
            The input data.
        z: torch.Tensor, optional
            The latent variable z scores for the provided data. If not provided, they will be computed using z_estimation_method.
        z_estimation_method : str, optional
            Method used to obtain the z scores. Can be 'NN', 'ML', 'EAP' or 'MAP' for neural network, maximum likelihood, expected a posteriori or maximum a posteriori respectively.
        average_over: str = "none", optional
            Whether to average the residuals and over which level. Can be 'everything', 'items', 'respondents' or 'none'. Use 'none' for no average. For example, with 'respondent' the residuals are averaged over all respondents and is thus an average per item. (default is 'none')
            
        Returns
        -------
        torch.Tensor
            The residuals.
        """
        data, z = self._evaluate_data_z_input(data, z, z_estimation_method)

        # 3D tensor with dimensions (respondents, items, item categories)
        if self.model.mc_correct is not None:
            probabilities = self.model.item_probabilities(z)
            # Creating a range tensor for slice indices
            respndents = torch.arange(probabilities.size(0)).view(-1, 1)
            # Expand slices to match the shape of indices
            expanded_respondents = respndents.expand_as(data)
            model_probs = probabilities[expanded_respondents, torch.arange(probabilities.size(1)), data.int()]
            residuals = 1 - model_probs
        else:
            residuals = data - self.model.expected_item_sum_score(z, return_item_scores=True)

        if average_over == "items":
            return residuals.mean(dim=1)
        if average_over == "respondents":
            return residuals.mean(dim=0)
        if average_over == "everything":
            return residuals.mean(dim=None)

        return residuals

    @torch.inference_mode()
    def group_fit_residuals(
        self,
        data: torch.Tensor = None,
        z: torch.Tensor = None,
        scale: str = "z",
        latent_variable: int = 1,
        standardize: bool = True,
        groups: int = 10,
        z_estimation_method: str = "ML",
        population_z: torch.Tensor = None,
        **kwargs
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Group the respondents based on their ordered latent variable scores.
        Calculate the residuals between the model estimated and observed data within each group.
        See ch. 20 in Handbook of Item Response Theory, Volume Two: Statistical Tools for more details.

        If 'data' is not supplied, the function defaults to using the model's training data.

        Parameters
        ----------
        data : torch.Tensor, optional
            A 2D tensor containing test data. Each row corresponds to one respondent and each column represents a latent variable. (default is None)
        z : torch.Tensor, optional
            A 2D tensor containing the pre-estimated z scores for each respondent in the data. Each row corresponds to one respondent and each column represents a latent variable. (default is None)
        scale : str, optional
            The grouping method scale, which can either be 'bit' or 'z'. Note: for uni-dimensional
            models, 'z' and 'bit' are equivalent. (default is 'z')
        latent_variable: int, optional
            Specifies the latent variable along which ordering and grouping should be performed. (default is 1)
        standardize : bool, optional
            Specifies whether the residuals should be standardized. (default is True)
        groups: int
            The number of groups. (default is 10)
        z_estimation_method : str, optional
            Method used to obtain the z scores. Can be 'NN', 'ML', 'EAP' or 'MAP' for neural network, maximum likelihood, expected a posteriori or maximum a posteriori respectively. (default is 'NN')
        population_z : torch.Tensor, optional
            Only for bit scores. The latent variable z scores for the population. If not provided, they will be computed using z_estimation_method with the model training data. (default is None)
        **kwargs : dict, optional
            Additional keyword arguments to be passed to the bit_scores_from_z method if scale is 'bit'. See :meth:`bit_scores_from_z` for details.
            
        Returns
        -------
        tuple[torch.Tensor, torch.Tensor]
            A tuple with torch tensors. The first one holds the residuals for each group and has dimensions (groups, items, item categories). The second one is a 1D tensor and holds the mid points of the groups.
        """
        grouped_data_probabilties, grouped_model_probabilties, group_mid_points = \
            self.latent_group_probabilities(
                data=data,
                z=z,
                scale=scale,
                latent_variable=latent_variable,
                groups=groups,
                z_estimation_method=z_estimation_method,
                population_z=population_z,
                **kwargs
            )

        raw_residuals = grouped_data_probabilties - grouped_model_probabilties
        if standardize:
            # get number of people in each group
            data = torch.chunk(data, groups)
            group_n = torch.tensor([group.shape[0] for group in data])

            std_residuals = raw_residuals / (grouped_model_probabilties * (1 - grouped_model_probabilties) / group_n.reshape(-1, 1, 1)).sqrt()
            return std_residuals, group_mid_points
        
        for item, categories in enumerate(self.model.item_categories):
            # set non-existing categories to nan
            categories = categories + 1 if self.model.model_missing else categories
            raw_residuals[:, item, categories:] = float("nan")

        return raw_residuals, group_mid_points
    
    @torch.inference_mode()
    def accuracy(
        self,
        data: torch.Tensor = None,
        z: torch.Tensor = None,
        z_estimation_method: str = "ML",
        level: str = "all",
    ):
        """
        Calculate the prediction accuracy of the model for the supplied data.

        Parameters
        ----------
        data : torch.Tensor
            The input data.
        z: torch.Tensor, optional
            The latent variable z scores for the provided data. If not provided, they will be computed using z_estimation_method.
        z_estimation_method : str, optional
            Method used to obtain the z scores. Can be 'NN', 'ML', 'EAP' or 'MAP' for neural network, maximum likelihood, expected a posteriori or maximum a posteriori respectively.
        level: str = "all", optional
            Specifies the level at which the accuracy is calculated. Can be 'all', 'item' or 'respondent'. For example, for 'item' the accuracy is calculated for each item. (default is 'all')

        Returns
        -------
        torch.Tensor
            The accuracy.
        """
        data, z = self._evaluate_data_z_input(data, z, z_estimation_method)

        probabilities = self.model.item_probabilities(z)
        accuracy = (torch.argmax(probabilities, dim=2) == data).float()

        if level == "item":
            dim = 0
        elif level == "respondent":
            dim = 1
        else:
            dim = None
        
        return accuracy.mean(dim=dim)

    @torch.inference_mode()
    def log_likelihood(
        self,
        data: torch.Tensor = None,
        z: torch.Tensor = None,
        z_estimation_method: str = "ML",
        reduction: str = "sum",
        level: str = "all",
    ):
        """
        Calculate the log-likelihood for the provided data.

        If 'data' is not supplied, the function defaults to using the model's training data.

        Parameters
        ----------
        data : torch.Tensor, optional
            A 2D tensor containing test data. Each row corresponds to one respondent and each column represents a latent variable. (default is None)
        z : torch.Tensor, optional
            A 2D tensor containing latent variable z scores. Each row corresponds to one respondent and each column represents a latent variable. (default is None)
        z_estimation_method : str, optional
            Method used to obtain the z scores. Can be 'NN', 'ML', 'EAP' or 'MAP' for neural network, maximum likelihood, expected a posteriori or maximum a posteriori respectively. (default is 'NN')
        reduction : str, optional
            Specifies the reduction method for the log-likelihood. Can be 'sum', 'none' or 'mean'. (default is 'sum')
        level : str, optional
            For reductions other than 'none', specifies the level at which the log-likelihood is summed/averaged. Can be 'all', 'item' or 'respondent'. For example, for 'item' the log-likelihood is summed/averaged for each item. (default is 'all')
            
        Returns
        -------
        torch.Tensor
            The log-likelihood for the provided data.
        """
        data, z = self._evaluate_data_z_input(data, z, z_estimation_method)

        if reduction != "none":
            if level == "item":
                dim = 0
            elif level == "respondent":
                dim = 1
            else:
                dim = None

        likelihoods = self.model.log_likelihood(
            data,
            self.model(z),
            loss_reduction="none"
        )
        
        if reduction in "mean":
            return likelihoods.view(z.shape[0], -1).mean(dim=dim)
        if reduction == "sum":
            return likelihoods.view(z.shape[0], -1).sum(dim=dim)
        
        return likelihoods

    def information(self, z: torch.Tensor, item: bool = True, scale: str = "z", degrees: list[int] = None, **kwargs) -> torch.Tensor:
        """
        Calculate the Fisher information matrix (FIM) for the z scores (or the information in the direction supplied by degrees).

        Parameters
        ----------
        z : torch.Tensor
            A 2D tensor containing latent variable z scores for which to compute the information. Each column represents one latent variable.
        item : bool, optional
            Whether to compute the information for each item (True) or for the test as a whole (False). (default is True)
        scale: str, optional
            The grouping method scale, which can either be 'bit' or 'z'. (default is 'z')
        degrees : list[int], optional
            For multidimensional models. A list of angles in degrees between 0 and 90, one for each latent variable. Specifies the direction in which to compute the information. (default is None)
        **kwargs : dict, optional
            Additional keyword arguments to be passed to the bit_score_gradients method if scale is 'bit'. See :meth:`bit_score_gradients` for details.
            
        Returns
        -------
        torch.Tensor
            A tensor with the information for each z score. Dimensions are:
            
            - By default: (z rows, items, FIM rows, FIM columns).
            - If degrees are specified: (z rows, items).
            - If item is False: (z rows, FIM rows, FIM columns).
            - If degrees are specified and item is False: (z rows).

        Notes
        -----
        In the context of IRT, the Fisher information matrix measures the amount of information
        that a test taker's responses :math:`X` carries about the latent variable(s)
        :math:`\\mathbf{z}`.

        The formula for the Fisher information matrix in the case of multiple parameters is:

        .. math::

            I(\\mathbf{z}) = E\\left[ \\left(\\frac{\\partial \\ell(X; \\mathbf{z})}{\\partial \\mathbf{z}}\\right) \\left(\\frac{\\partial \\ell(X; \\mathbf{z})}{\\partial \\mathbf{z}}\\right)^T \\right] = -E\\left[\\frac{\\partial^2 \\ell(X; \\mathbf{z})}{\\partial \\mathbf{z} \\partial \\mathbf{z}^T}\\right]

        Where:

        - :math:`I(\\mathbf{z})` is the Fisher Information Matrix.
        - :math:`\ell(X; \\mathbf{z})` is the log-likelihood of :math:`X`, given the latent variable vector :math:`\\mathbf{z}`.
        - :math:`\\frac{\\partial \\ell(X; \\mathbf{z})}{\\partial \\mathbf{z}}` is the gradient vector of the first derivatives of the log-likelihood of :math:`X` with respect to :math:`\\mathbf{z}`.
        - :math:`\\frac{\\partial^2 \\log f(X; \\mathbf{z})}{\\partial \\mathbf{z} \\partial \\mathbf{z}^T}` is the Hessian matrix of the second derivatives of the log-likelihood of :math:`X` with respect to :math:`\\mathbf{z}`.
        
        For additional details, see :cite:t:`Chang2017`.
        """
        if degrees is not None and len(degrees) != self.model.latent_variables:
            raise ValueError("There must be one degree for each latent variable.")

        probabilities = self.model.item_probabilities(z.clone())
        gradients = self.model.probability_gradients(z).detach()
        if scale == "bit":
            bit_z_gradients = self.scorer.bit_score_gradients(z, one_dimensional=False, **kwargs)
            # we divide by diagonal of the bit score gradients (the gradients in the direction of the bit score corresponding z scores)
            bit_z_gradients_diag = torch.einsum('...ii->...i', bit_z_gradients)
            gradients = (gradients.permute(1, 2, 0, 3) / bit_z_gradients_diag).permute(2, 0, 1, 3)

        # squared gradient matrices for each latent variable
        # Uses einstein summation with batch permutation ...
        squared_grad_matrices = torch.einsum("...i,...j->...ij", gradients, gradients)
        information_matrices = squared_grad_matrices / probabilities.unsqueeze(-1).unsqueeze(-1).expand_as(squared_grad_matrices)
        information_matrices = information_matrices.nansum(dim=2) # sum over item categories

        if degrees is not None and z.shape[1] > 1:
            cos_degrees = torch.tensor(degrees).float().deg2rad_().cos_()
            # For each z and item: Matrix multiplication cos_degrees^T @ information_matrix @ cos_degrees
            information = torch.einsum('i,...ij,j->...', cos_degrees, information_matrices, cos_degrees)
        else:
            information = information_matrices
            if self.model.latent_variables == 1:
                information.squeeze_()

        if item:
            return information
        else:
            return information.nansum(dim=1) # sum over items


    @torch.inference_mode()
    def group_fit_log_likelihood(
        self,
        data: torch.Tensor = None,
        z: torch.Tensor = None,
        z_estimation_method: str = "ML",
        groups: int = 10,
        latent_variable: int = 1,
    ):
        """
        Group the respondents based on their ordered latent variable scores.
        Calculate the average log-likelihood of the data within each group.

        If 'data' is not supplied, the function defaults to using the model's training data.

        Parameters
        ----------
        data : torch.Tensor, optional
            A 2D tensor containing test data. Each row corresponds to one respondent and each column represents a latent variable. (default is None)
        z : torch.Tensor, optional
            A 2D tensor containing the pre-estimated z scores for each respondent in the data. Each row corresponds to one respondent and each column represents a latent variable. (default is None)
        z_estimation_method : str, optional
            Method used to obtain the z scores. Can be 'NN', 'ML', 'EAP' or 'MAP' for neural network, maximum likelihood, expected a posteriori or maximum a posteriori respectively. (default is 'NN')
        groups: int
            The number of groups. (default is 10)
        latent_variable: int, optional
            Specifies the latent variable along which ordering and grouping should be performed. (default is 1)

        Returns
        -------
        torch.Tensor
            The average log-likelihood for each group.
        """
        data, z = self._evaluate_data_z_input(data, z, z_estimation_method)

        indicies = torch.sort(z[:, latent_variable - 1], dim=0)[1]
        z = z[indicies]
        data = data[indicies]
        likelihoods = self.model.log_likelihood(
            data,
            self.model(z),
            loss_reduction="none"
        )
        respondent_likelihoods = likelihoods.view(z.shape[0], -1).sum(dim=1)
        group_mean_likelihoods = torch.zeros(groups)
        start_index = 0
        for group in range(groups):
            group_size = respondent_likelihoods.shape[0] // groups
            remainder = len(respondent_likelihoods) % groups
            if group < remainder:
                group_size += 1
            group_mean_likelihoods[group] = respondent_likelihoods[start_index: start_index + group_size].mean()
            start_index += group_size
            
        return group_mean_likelihoods

    def latent_group_probabilities(
        self,
        data: torch.Tensor = None,
        z: torch.Tensor = None,
        scale: str = "z",
        latent_variable: int = 1,
        groups: int = 10,
        z_estimation_method: str = "ML",
        population_z: torch.Tensor = None,
        **kwargs
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Group the respondents based on their ordered latent variable scores.
        Calculate both the observed and IRT model probabilities for each possible item response, within each group.

        If 'data' is not supplied, the function defaults to using the model's training data.

        Parameters
        ----------
        data : torch.Tensor, optional
        z : torch.Tensor, optional
            The latent variable z scores for the provided data. If not provided, they will be computed using z_estimation_method. (default is None)
        scale : str, optional
            The grouping method scale, which can either be 'bit' or 'z'. Note: for uni-dimensional
            models, 'z' and 'bit' are equivalent. (default is 'z')
        latent_variable: int, optional
            Specifies the latent variable along which ordering and grouping should be performed. (default is 1)
        groups: int
            The number of groups. (default is 10)
        z_estimation_method : str, optional
            Method used to obtain the z scores. Can be 'NN', 'ML', 'EAP' or 'MAP' for neural network, maximum likelihood, expected a posteriori or maximum a posteriori respectively. (default is 'NN')
            A 2D tensor containing test data. Each row corresponds to one respondent and each column represents a latent variable. (default is None)
        population_z : torch.Tensor, optional
            Only for bit scores. The latent variable z scores for the population. If not provided, they will be computed using z_estimation_method with the model training data. (default is None)
        **kwargs : dict, optional
            Additional keyword arguments to be passed to the latent_scores method if scores_to_plot is not provided. See :meth:`latent_scores` for details.

        Returns
        -------
        tuple[torch.Tensor, torch.Tensor, torch.Tensor]
            A 3D torch tensor with data group averages. The first dimension represents the groups, the second dimension represents the items and the third dimension represents the item categories.

            A 3D torch tensor with model group averages. The first dimension represents the groups, the second dimension represents the items and the third dimension represents the item categories.

            The third tensor contains the average latent variable values within each group along the specified latent_variable.
        """

        if scale not in ["bit", "z"]:
            raise ValueError("Invalid scale. Choose either 'z' or 'bit'.")

        data, z = self._evaluate_data_z_input(data, z, z_estimation_method)

        if scale == "bit":
            if population_z is None and data is self.algorithm.train_data:
                population_z = z
            
            bit_scores, _ = self.scorer.bit_scores_from_z(
                z=z,
                one_dimensional=False,
                z_estimation_method=z_estimation_method,
                **kwargs
            )

            # Sort based on correct column and get the sorted indices
            _, indices = torch.sort(
                bit_scores[:, latent_variable - 1], dim=0
            )
            # Use the indices to sort
            bit_scores = bit_scores[indices]
            z = z[indices]

            grouped_bit = torch.chunk(bit_scores, groups)
            grouped_z = torch.chunk(z, groups)

            group_mid_points = torch.tensor(
                [
                    group[:, latent_variable - 1].mean()
                    for group in grouped_bit
                ]
            )
        elif scale == "z":
            _, indices = torch.sort(z[:, latent_variable - 1], dim=0)
            z = z[indices]
            grouped_z = torch.chunk(z, groups)
            group_mid_points = torch.tensor(
                [group[:, latent_variable - 1].median() for group in grouped_z]
            )

        data = data[indices]
        grouped_data = torch.chunk(data, groups)
        grouped_data_probabilties = self._grouped_data_probabilities(grouped_data)
        grouped_model_probabilties = self._grouped_z_probabilities(grouped_z)
        return grouped_data_probabilties, grouped_model_probabilties, group_mid_points

    @torch.inference_mode()
    def _grouped_z_probabilities(self, grouped_z: tuple[torch.Tensor, ...]):
        """
        Computes the average probabilities for each potential item response for each group.

        Parameters
        ----------
        grouped_z : tuple[torch.Tensor, ...]
            A tuple containing 2D tensors. Each tensor represents a group of respondents, with the first dimension corresponding to the respondents and the second dimension representing their latent variables in the form of z-scores.

        Returns
        -------
        torch.Tensor
            A 3D torch tensor with group averages. The first dimension represents the groups, the second dimension represents the items and the third dimension represents the item categories.
        """
        group_probabilities = torch.zeros(len(grouped_z), len(self.model.modeled_item_responses), max(self.model.modeled_item_responses))
        for group_i, group in enumerate(grouped_z):
            item_probabilities = self.model.item_probabilities(group)
            group_probabilities[group_i, :, :] = item_probabilities.mean(dim=0)
        return group_probabilities

    @torch.inference_mode()
    def _grouped_data_probabilities(self, grouped_data: tuple[torch.Tensor, ...]):
        """
        Computes the average probabilities for each potential item response for each group.

        Parameters
        ----------
        grouped_data : tuple[torch.Tensor, ...]
            A tuple containing 2D tensors. Each tensor represents a group of respondents, with the first dimension representing the respondents and the second dimension representing items.

        Returns
        -------
        torch.Tensor
            A 3D torch tensor with group averages. The first dimension represents the groups, the second dimension represents the items and the third dimension represents the item categories.
        """
        modeled_item_responses = self.model.modeled_item_responses
        group_probabilities = torch.zeros(len(grouped_data), len(modeled_item_responses), max(modeled_item_responses))
        for group_i, group in enumerate(grouped_data):
            for item_i, _ in enumerate(modeled_item_responses):
                counts = torch.bincount(group[:, item_i].int(), minlength=max(modeled_item_responses))
                group_probabilities[group_i, item_i, :] = counts.float() / counts.sum()

        return group_probabilities

    @torch.inference_mode()
    def sum_score_probabilities(
        self,
        latent_density_method: str = "data",
        population_data: torch.Tensor = None,
        trapezoidal_segments: int = 1000,
        sample_size: int = 100000,
    ):
        """
        Computes the marginal probabilities for each sum score, averged over the latent space density. For 'qmvn' and 'gmm' densities, the trapezoidal rule is used for integral approximation.

        Parameters
        ----------
        latent_density_method : str, optional
            Specifies the method used to approximate the latent space density.
            Possible options are
            - 'data' averages over the z scores from the population data.
            - 'encoder sampling' samples z scores from the encoder. Only available for VariationalAutoencoderIRT models
            - 'qmvn' for quantile multivariate normal approximation of a multivariate joint density function (QuantileMVNormal class).
            - 'gmm' for an sklearn gaussian mixture model.
        population_data : torch.Tensor, optional
            The population data used for approximating sum score probabilities. Default is None and uses the training data.
        trapezoidal_segments : int, optional
            The number of integration approximation intervals for each z dimension. (Default is 1000)
        sample_size : int, optional
            Sample size for the 'encoder sampling' method. (Default is 100000)
        Returns
        -------
        torch.Tensor
            A 1D tensor with the probability for each total score.
        """
        self._validate_latent_density_method(latent_density_method)
        z_scores, weights = self._get_z_scores_and_weights(
            latent_density_method, population_data, trapezoidal_segments, sample_size
        )
        probabilities = self.model.item_probabilities(z_scores)
        # sum incorrect response option probabilities if MC
        if self.model.mc_correct is not None:
            probabilities = sum_incorrect_probabilities(
                probabilities=probabilities,
                modeled_item_responses=self.model.modeled_item_responses,
                mc_correct=self.model.mc_correct,
                missing_modeled=self.model.model_missing
            )
            item_categories = [2] * len(self.model.modeled_item_responses)
        else:
            item_categories = self.model.item_categories
            # Add together probabilities for missing response and a score of 0
            if self.model.model_missing:
                summed_slices = probabilities[:, :, 0] + probabilities[:, :, 1]
                probabilities = torch.cat((summed_slices.unsqueeze(-1), probabilities[:, :, 2:]), dim=2)
        
        conditional_total_score_probs = conditional_score_distribution(
            probabilities, item_categories
        )
        sum_score_probabilities = conditional_total_score_probs * weights.view(-1, 1)
        return sum_score_probabilities.sum(dim=0)

    def _validate_latent_density_method(self, latent_density_method: str) -> None:
        valid_methods = ["data", "encoder sampling", "qmvn", "gmm"]
        if latent_density_method not in valid_methods:
            raise ValueError(
                f"Invalid latent density method. Must be one of {valid_methods}."
            )
        if latent_density_method == "encoder sampling" and not isinstance(
            self.algorithm, VAEIRT
        ):
            raise ValueError(
                "Encoder sampling is only available for variational autoencoder models."
            )

    def _get_z_scores_and_weights(
        self,
        latent_density_method: str,
        population_data: torch.Tensor,
        trapezoidal_segments: int,
        sample_size: int,
        ml_map_device: str = "cuda" if torch.cuda.is_available() else "cpu",
        lbfgs_learning_rate: float = 0.3,
    ):
        if population_data is None:
            z_scores = self.algorithm.training_z_scores
        else:
            z_scores = self.scorer.latent_scores(population_data, scale="z", z_estimation_method="NN", ml_map_device=ml_map_device, lbfgs_learning_rate=lbfgs_learning_rate)

        if latent_density_method in ["data", "encoder sampling"]:
            weights = (
                torch.ones(z_scores.shape[0])
                / z_scores.shape[0]
            )
            if latent_density_method == "encoder sampling":
                z_scores = self.algorithm.sample_latent_variables(
                    sample_size=sample_size, input_data=population_data
                )
        else:
            (
                z_scores,
                weights,
            ) = self._get_z_scores_and_weights_for_latent_density_methods(
                latent_density_method, z_scores, population_data, trapezoidal_segments
            )

        return z_scores, weights

    def _get_z_scores_and_weights_for_latent_density_methods(
        self,
        latent_density_method: str,
        z_scores: torch.Tensor,
        population_data: torch.Tensor,
        trapezoidal_segments: int,
    ):
        # We approximate the density if
        # population_data is not none and the correct density
        # is not already in self.scorer.latent_density
        if population_data is not None or (
            (
                latent_density_method != "qmvn"
                or not isinstance(self.scorer.latent_density, QuantileMVNormal)
            )
            and (
                latent_density_method != "gmm"
                or not isinstance(self.scorer.latent_density, GaussianMixtureTorch)
            )
        ):
            self.scorer.approximate_latent_density(
                z_scores=z_scores, approximation=latent_density_method
            )

        # get the min/max points for integration
        min_z, max_z = self.scorer.min_max_z_for_integration(z_scores)

        # Create a list of linspace tensors for each dimension
        lin_spaces = [
            torch.linspace(min_z[i], max_z[i], trapezoidal_segments)
            for i in range(len(min_z))
        ]

        # Use torch.cartesian_prod to generate all combinations
        z_scores = torch.cartesian_prod(*lin_spaces)

        if z_scores.dim() == 1:
            # Add an extra dimension for 1D models to make it a 2D tensor with 1 column
            z_scores = z_scores.unsqueeze(1)

        weights = self.scorer.latent_density.pdf(z_scores)
        weights = weights / weights.sum()

        return z_scores, weights
